# $Id: logger.tcl,v 1.1 2002/11/02 20:42:02 alexey Exp $



namespace eval ::logger {
    variable logdir ~/.tkabber/logs

    if {![file exists $logdir]} {
	file mkdir $logdir
    }
}

proc ::logger::log_file {chatid} {
    variable logdir

    regsub -all @|/ $chatid _ filename 
    return ${logdir}/$filename
}

proc ::logger::log_message {chatid from type body x} {
    set nas [node_and_server_from_jid $chatid]
    if {$type == "chat" && ![chat::is_groupchat $nas]} {
	set chatid $nas
    }

    set logfile [log_file $chatid]
    set nick [chat::get_nick $from $type]

    set seconds [clock seconds]
    foreach xelem $x {
	jlib::wrapper:splitxml $xelem tag vars isempty chdata children
	
	if {[cequal [jlib::wrapper:getattr $vars xmlns] jabber:x:delay]} {
	    set seconds [clock scan [jlib::wrapper:getattr $vars stamp] -gmt 1]
	}
    }
    set ts [clock format $seconds -format "%Y%m%dT%H%M%S"]

    set fd [open $logfile a]
    fconfigure $fd -encoding utf-8
    puts $fd [list [list timestamp $ts nick ${nick} body $body]]
    close $fd
}

hook::add draw_message_hook ::logger::log_message 15


#set menuhistidx [expr [.jidpopupmenu index "Edit groups"] + 1]
#.jidpopupmenu insert $menuhistidx command -label "Show history" \
#    -command {logger::show_log $curuser}


proc ::logger::winid {name} {
    #regsub -all \\. $name | allowed_name 
    set allowed_name [jid_to_tag $name]
    return .log_$allowed_name
}


proc ::logger::show_log {jid} {
    global font
    global tcl_platform
    global defaultnick
    global loginconf

    set nas [node_and_server_from_jid $jid]
    if {![chat::is_groupchat $nas]} {
	set jid $nas
    }

    set logfile [log_file $jid]
    if {[file exists $logfile]} {
	set lw [winid $jid]
	debugmsg plugins "LOGGER: $lw"
	if {[winfo exists $lw]} {
	    focus -force $lw
	    return
	}

	if { [string equal $tcl_platform(platform) "unix"] } {
	    set re raised
	    set bd 1
	} else {
	    set re flat
	    set bd 0
	}


	toplevel $lw -relief $re -borderwidth $bd -class Chat
	wm withdraw $lw
	wm title $lw "History for $jid"

	set bbox [ButtonBox $lw.bbox -spacing 10 -padx 10 -default 0]
	$bbox add -text "Close" -command [list destroy $lw]
	$bbox setfocus 0
	pack $bbox -padx 2m -pady 2m -anchor e -side bottom
	bind $lw <Escape> "ButtonBox::invoke $bbox default"
	bind $lw <Return> "ButtonBox::invoke $bbox default"

	set sep [Separator::create $lw.sep -orient horizontal]
	pack $sep -pady 1m -fill x -side bottom

	set lf [ScrolledWindow $lw.sw]
	pack $lf -padx 2m -pady 2m

	text $lf.log -font $font -wrap word -takefocus 1
	set l $lf.log
	pack $lf.log
	$lf setwidget $l
	focus $l

	#debugmsg plugins "LOGGER: '[option get $lf theyforeground Chat]'"
	$l tag configure they -foreground [option get $lw theyforeground Chat]
	$l tag configure me -foreground [option get $lw meforeground Chat]
	$l tag configure server_lab \
	    -foreground [option get $lw serverlabelforeground Chat]
	$l tag configure server \
	    -foreground [option get $lw serverforeground Chat]

	set fd [open $logfile r]
	fconfigure $fd -encoding utf-8
	set hist [read $fd]
	close $fd

	if {[info exists defaultnick($jid)]} {
	    set mynick $defaultnick($jid)
	} else {
	    set mynick $loginconf(user)
	}

	foreach vars $hist {
	    array unset tmp
	    array set tmp $vars
	    if {[info exists tmp(timestamp)]} {
		set seconds [clock scan $tmp(timestamp) -gmt 0]
		$l insert end [clock format $seconds -format {[%Y-%m-%d %X]}]
		#$l insert end \[$tmp(timestamp)\]
	    }
	    if {[info exists tmp(nick)] && $tmp(nick) != ""} {
		if {$tmp(nick) == $mynick} {
		    set tag me
		} else {
		    set tag they
		}
		if {[info exists tmp(body)] && [regsub {^/me } $tmp(body) {} body]} {
		    $l insert end "*$tmp(nick) $body" $tag
		    unset tmp(body)
		} else {
		    $l insert end "<$tmp(nick)>" $tag
		}
		set servertag ""
	    } else {
		$l insert end "---" server_lab
		set servertag server
	    }
	    if {[info exists tmp(body)]} {
		$l insert end " $tmp(body)" $servertag
	    }
	    if {![$l compare "end -1 chars linestart" == "end -1 chars"]} {
		$l insert end "\n"
	    }
	}

	$lf.log see end
	$lf.log configure -state disabled

	wm deiconify $lw
	
    } else {
	
    }
}

